function CreateFigure5ab
%Create figures 5a and 5b

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%INITIALIZATION%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
numFrames = 200;
cellBoundary = 10;
numDimensions = 1;

%Create a set of particles that define our prior belief
%We need to make sure the true models aren't too fast for this simulation, because skipping frames only makes
%sense if objects are moving slowly (relative to the pixel resolution)...if
%we had fast models, then the intelligent algorithm would just choose to
%acquire every frame
sigmaLowHigh = [0 1];
muLowHigh = [-0.5 0.5];
numPriorParticles = 1000;
priorParticles = rand(numPriorParticles, 2*numDimensions+1)*(sigmaLowHigh(2) - sigmaLowHigh(1)) + sigmaLowHigh(1);
priorParticles(:,(numDimensions+1):(2*numDimensions)) = rand(size(priorParticles(:,(numDimensions+1):(2*numDimensions))))*(muLowHigh(2) - muLowHigh(1)) + muLowHigh(1);
priorParticles(:,2*numDimensions+1) = 1/numPriorParticles;


%Set up the pixel structure of the image, using different numbers and sizes
%of pixels in every frame
resolution = 1;
for i = 1:numFrames
    pixelBoundary{i} = 0:resolution:cellBoundary;
end

%The intelligent acquisiiton algorithm works best with 100,000+ particles,
%but we can reduce it to 10,000 for a faster result...
%Because these results can be noisy, it may be worth increasing the number
%of trajectories (the more the better) if time allows.
numParticles = 100000;
numTrajectories = 100;
parameters.YData{1} = 0;
parameters.YData{2} = 0;
parameters.YData{3} = 0;
photobleaching{1} = 0;
photobleaching{2} = 0;
photobleaching{3} = 0;
totalCaptured = 0;
numObjects = 1;

%Create trajectories
for(iTrajectory=1:numTrajectories)
    for(iObject = 1:numObjects)
        sigma{iTrajectory, iObject} = rand(1,1)*(sigmaLowHigh(2) - sigmaLowHigh(1))+sigmaLowHigh(1);    
        mu{iTrajectory, iObject} = rand(1,1)*(muLowHigh(2) - muLowHigh(1))+muLowHigh(1);  
        position{iTrajectory, iObject} = CreateTrajectory(length(sigma{iTrajectory,iObject}), cellBoundary,numFrames,sigma{iTrajectory,iObject},mu{iTrajectory,iObject},5.5);
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%MAIN CODE%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%Loop through trials acquiring a frame t+1 only when the
%benefit of acquiring frame t+2 exceeds the benefit of
%acquiring frame t+1.
intelligentAcquisition = InitializeIntelligentAcquisition();
intelligentAcquisition.numFramePeeks = 2;
intelligentAcquisition.whenToAcquire = 2;
intelligentAcquisition.photobleachingCost = 1;
intelligentAcquisition.timeCost = 0.1;

for(iTrajectory = 1:numTrajectories)
    iTrajectory
    for iObject = 1:numObjects
        pos{iObject} = position{iTrajectory,iObject};
        truth{iObject} = [sigma{iTrajectory,iObject} mu{iTrajectory,iObject}];
    end
    [predictions{1},uncertainty,estUncertainty{1}, futureUncertainty, particle, captured{1}, output{1}] = LearnModel([1 1 1 0],pos, cellBoundary, numParticles, pixelBoundary, truth, intelligentAcquisition, priorParticles);
    parameters.XData{1} = 1:numFrames;
    parameters.YData{1} = parameters.YData{1} + uncertainty;
    
    %Photobleaching measured by the number of object exposures
    captured{1}((end+1):numFrames)=0;    
    photobleaching{1} = photobleaching{1}+cumsum(captured{1});  
    
    %Record the total number of times a frame was captured (used to set the
    %frame rate for the control algorithm later)
    totalCaptured = totalCaptured + sum(sum(captured{1}));
end

%Loop through trials acquiring 1) at a constant frame rate such that the total
%number of frame acquisitions is the same as for the previous acquisition
%method, 2) every frame
for(iTrajectory = 1:numTrajectories)
    iTrajectory
    
    %Learn the model of this trajectory when acquiring at a constant frame
    %rate equal to the average frame rate from the previous acquisition
    %method
    intelligentAcquisition.numFramePeeks = 1;
    intelligentAcquisition.whenToAcquire = 0;
    intelligentAcquisition.skipFrame = numFrames*numTrajectories*numObjects/totalCaptured;
    for iObject = 1:numObjects
        pos{iObject} = position{iTrajectory,iObject};
        truth{iObject} = [sigma{iTrajectory,iObject} mu{iTrajectory,iObject}];
    end
    [predictions{2},uncertainty,estUncertainty{2}, futureUncertainty, particle, captured{2},output{2}] = LearnModel([1 1 1 0],pos, cellBoundary, numParticles, pixelBoundary, truth, intelligentAcquisition, priorParticles);
    parameters.XData{2} = 1:numFrames;
    parameters.YData{2} = parameters.YData{2} + uncertainty;
    
    captured{2}((end+1):numFrames)=0;
    photobleaching{2} = photobleaching{2}+cumsum(captured{2});

    %Learn the model of this trajectory when acquiring every frame
    intelligentAcquisition.whenToAcquire = -1;
    [predictions{3},uncertainty,estUncertainty{3}, futureUncertainty, particle, captured{3},output{3}] = LearnModel([1 1 1 0],pos, cellBoundary, numParticles, pixelBoundary, truth, intelligentAcquisition, priorParticles);
    parameters.XData{3} = 1:numFrames;
    parameters.YData{3} = parameters.YData{3} + uncertainty;
    
    captured{3}((end+1):numFrames)=0;
    photobleaching{3} = photobleaching{3}+cumsum(captured{3});

end

%Normalize results
normFactor = parameters.YData{3}(1);
parameters.YData{1} = parameters.YData{1}/normFactor;
parameters.YData{2} = parameters.YData{2}/normFactor;
parameters.YData{3} = parameters.YData{3}/normFactor;
parameters.YRange{1} = [0 1];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%SHOW RESULTS%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

parameters.XLabels{1} = 'Frame number';
parameters.YLabels{1} = 'Error';
parameters.colorStyle{1} = 'r';
parameters.colorStyle{2} = 'b';  
parameters.lineStyle{2} = '--';
parameters.colorStyle{3} = 'k';
parameters.lineStyle{3} = '--';
parameters  = bimagic_plot(parameters);

parameters2 = parameters;
parameters2.XData{1} = photobleaching{1}/numTrajectories;
parameters2.XData{2} = photobleaching{2}/numTrajectories;
parameters2.XData{3} = photobleaching{3}/numTrajectories;

bimagic_plot(parameters);
bimagic_plot(parameters2);
a=1;
