%LearnModel
%Learn the model of a given trajectory:
%Input: include - binary vector specifying what quantities need to be caluclated
%       position - specifies the object coordinates in each frame
%       cellBoundary - the maximum coordinate in the cell
%       numParticles - the number of particles to use in the particle
%       filter
%       pixelBoundaries - the points at which the pixels start and stop
%       truth - the true model, used only for calculating ground truth
%       errors
%       intelligentAcquisition - a structure that specifies the acquisiiton
%       methods to use
%       priorParticles - a set of particles that dictates the prior
%       distribution of the model
%Output: prediction - a prediction of the object model at each frame
%       uncertainty - the true error/uncertainty at each frame
%       estUncertainty - the estimate of the error at each frame
%       futureUncertainty - the prediction of the error for the subsequent
%       frame
%       particle - the final set of particles in the particle filter
%       captured - binary vector indicating which frames resulted in an
%       object being acquired
%       output - structure contianing other information from the model
%       building process

function [prediction, uncertainty, estUncertainty, futureUncertainty, particle, captured, output] = LearnModel(include, position, cellBoundary, numParticles, pixelBoundaries, truth, intelligentAcquisition, priorParticles)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%INITIALIZE VARIABLES%%%%%%%%

%Determine number of objects based on number of trajectories in position
%cell array
numObjects = length(position);

%For constant frame rates
if(intelligentAcquisition.whenToAcquire == 0)
    acquireNow = 0;
end

%If using stopping criteria algorithms, we must calculate estUncertainty
%and futureUncertainty
if(intelligentAcquisition.stoppingCriteria > 0)
    include([3 4]) = 1;
end

%If number of frames not specified, use all the frames in the trajectory
numFrames = intelligentAcquisition.numFrames;
if(numFrames == -1)
    numFrames = size(position{1}, 1);    
end

%Intialize the current cost incurred at zero
output.timeCost = 0;
output.photobleachingCost = 0;

%Initialize the time elapsed since the last acquisition to zero
elapsedTime = 0;

%Set number of dimensions, based on the dimensions in the input
%trajectories
numDimensions = size(position{1},2);

%Specify which indices indicate the start of meta information in the
%particle and priorParticle arrays
particleMetaStart = 3*numDimensions;
priorMetaStart = 2*numDimensions;

%%%%%%%%SET UP PARTICLE FILTER%%%%%%%%

%Choose particles from the set of prior particles
priorIDs = randsample(1:size(priorParticles,1),numParticles,true,priorParticles(:,priorMetaStart+1));
particle{1} = rand(numParticles, particleMetaStart+1)*cellBoundary;
particle{1}(1:numParticles,(numDimensions+1):(3*numDimensions)) = priorParticles(priorIDs, 1:(2*numDimensions)); %added for prior
particle{1}(:,particleMetaStart+2)=priorIDs;
particle{1}(:,particleMetaStart+3)=1/numParticles;

%Duplicate particles for all objects, and set their object IDs, and
%initialize their positions to the first frame of the movie
particle{1} = repmat(particle{1},[numObjects 1]);
for(iObject = 1:numObjects)
    particle{1}(((iObject-1)*numParticles+1):(iObject*numParticles),1:numDimensions) = repmat(position{iObject}(1,1:numDimensions), [numParticles 1]);    
    particle{1}(((iObject-1)*numParticles+1):(iObject*numParticles),particleMetaStart+1)=iObject; 
end

modelIndices = (numDimensions+1):(3*numDimensions);
notModelIndices = [1:numDimensions (particleMetaStart+1):size(particle{1},2)];

%Compute initial uncertainty, estimated uncertainty, and prediction of the
%motion model for each object
for(iObject = 1:numObjects)
    indices = find(particle{1}(:,particleMetaStart+1)==iObject);
    prediction(1,iObject,1:2*numDimensions) = sum(particle{1}(indices,numDimensions+1:3*numDimensions).*repmat(particle{1}(indices,particleMetaStart+3),1,2*numDimensions),1)/sum(particle{1}(indices,particleMetaStart+3)); 
    uncertainty(1,iObject) = (sum(sum(repmat(particle{1}(indices,particleMetaStart+3),1,numDimensions).*(particle{1}(indices,numDimensions+1:2*numDimensions)-repmat(truth{iObject}(1:numDimensions), [length(indices) 1])).^2))/sum(particle{1}(indices,particleMetaStart+3))) + (sum(sum(repmat(particle{1}(indices,particleMetaStart+3),1,numDimensions).*(particle{1}(indices,2*numDimensions+1:3*numDimensions)-repmat(truth{iObject}((numDimensions+1):(2*numDimensions)), [length(indices) 1])).^2))/sum(particle{1}(indices,particleMetaStart+3)));
    estUncertainty(1,iObject) = 2*sum(var(particle{1}(indices,numDimensions+1:3*numDimensions)));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%MAIN CODE%%%%%%%%

%%%%%%%%ASSIGN METRICS%%%%%%%%
%Assign each particle a pixelID
pixel{1} = FindPixels(pixelBoundaries{1}, particle{1}, numDimensions); 

%Calculate the expected future uncertainty
if(include(4))
for(iObject = 1:numObjects)
indices = find(particle{1}(:,particleMetaStart+1)==iObject);  
[futureUncertainty(1,1,iObject)] = PredictUncertainty(particle{1}(indices,:), numDimensions,3*numDimensions,numDimensions, pixel{1}(indices,:));
end
else
    futureUncertainty = 0;
end


%We always capture objects in the first frame
for iObject = 1:numObjects
    captured(1,iObject) = 1;
end

numDetectedObjects = numObjects;

%%%%%%%%LOOP THROUGH FRAMES%%%%%%%%

for iFrame = 2:numFrames
    iFrame
    
    %Determine particle values in the next few frames
    for iFramePeek = 1:intelligentAcquisition.numFramePeeks
        %move particles
        if(iFramePeek > 1)
            particle{iFramePeek} = MoveParticles(particle{iFramePeek-1}, numDimensions, cellBoundary);
        else
            particle{1} = MoveParticles(particle{1}, numDimensions, cellBoundary);
        end
    end
    
    for iFramePeek = 1:intelligentAcquisition.numFramePeeks
        pixel{iFramePeek} = FindPixels(pixelBoundaries{iFrame}, particle{iFramePeek}, numDimensions); 
  
        %%%%%%%%WHERE TO ACQUIRE%%%%%%%%
        
        %These methods currently only work for the single object case
        switch intelligentAcquisition.whereToAcquire
            %Choose pixels with highest reward
            case 1
                totalBenefit = 0;
                maxPixel = max(pixel{iFramePeek}(:,numDimensions+1));
                for iObject = 1:numObjects
                    indices = find(particle{iFramePeek}(:,particleMetaStart+1)==iObject);
                    benefit = FindPixelReward(pixel{iFramePeek}(indices,:), numDimensions, particle{iFramePeek}(indices,:), estUncertainty(iFrame-1), [],[],numDimensions+1,3*numDimensions,ones(2*numDimensions,1),maxPixel);
                    benefit(isnan(benefit)) = 0;
                    totalBenefit = totalBenefit + benefit;
                end
                [temp order] = sort(totalBenefit);
                chosenPixels = order(max(1,end-intelligentAcquisition.numPixels+1):end);
            
            %Choose pixels most likely to contain object
            case 2
                totalBenefit = 0;
                maxPixel = max(pixel{iFramePeek}(:,numDimensions+1));
                for iObject = 1:numObjects
                    indices = find(particle{iFramePeek}(:,particleMetaStart+1)==iObject);
                    benefit = FindPixelProbability(pixel{iFramePeek}(indices,:), numDimensions,maxPixel);
                    benefit(isnan(benefit)) = 0;
                    totalBenefit = totalBenefit + benefit;
                end
                [temp order] = sort(totalBenefit);
                chosenPixels = order(max(end-intelligentAcquisition.numPixels+1,1):end);
            
            %Choose pixels randomly
            case 3
                order = randperm((length(pixelBoundaries{iFrame})-1)^(numDimensions));
                chosenPixels = order(1:intelligentAcquisition.numPixels);
            
            %Choose pixels with highest benefit over next 2 frames
            case 4
                totalBenefit = 0;
                particle{iFramePeek+1} = MoveParticles(particle{iFramePeek}, numDimensions, cellBoundary);
                pixel{iFramePeek+1} = FindPixels(pixelBoundaries{iFrame}, particle{iFramePeek+1}, numDimensions); 
                numPixels = max(max(pixel{iFramePeek}(:,end)),max(pixel{iFramePeek+1}(:,end)));
                pixelTrajectory = CreatePixelTrajectory(pixel{iFramePeek},pixel{iFramePeek+1},numPixels);            
                for(iObject =1:numObjects)
                    indices = find(particle{iFramePeek}(:,particleMetaStart+1)==iObject);
                    benefit = FindPixelBenefit(pixelTrajectory(indices,:), numDimensions, particle{iFramePeek}(indices,:), estUncertainty(iFrame-1), [],[],numDimensions+1,3*numDimensions,ones(2*numDimensions,1),numPixels);
                    benefit = FindPixelBenefitFromTrajectory(benefit,pixelTrajectory(indices,:),numPixels);
                    benefit(isnan(benefit)) = 0;
                    totalBenefit = totalBenefit + benefit;
                end
                [temp order] = sort(totalBenefit);
                chosenPixels = order(end-intelligentAcquisition.numPixels+1:end);
            
            %Choose pre-defined set of pixels
            case 5
                chosenPixels = intelligentAcquisition.chosenPixels;
                
            %Choose pixels with highest reward/cost ratio (benefit)
            case 6
                %find the first pixel
                [reward pixelProperties] = FindPixelReward(pixel{iFramePeek}, numDimensions, particle{iFramePeek}, estUncertainty(iFrame-1), [],[],numDimensions+1,3*numDimensions,ones(2*numDimensions,1),-1);
                reward(isnan(reward)) = 0;
                cost = FindPixelProbability(pixel{iFramePeek}, numDimensions,-1);
                cost(isnan(reward)) = 0;
                cost = cost/sum(cost)*intelligentAcquisition.photobleachingCost;
                totalCost = cost + intelligentAcquisition.timeCost;
                rewardcostratio = reward ./ totalCost;
                [maxratio index] = max(rewardcostratio);
                chosenPixels = index;
                totalCost = totalCost + cost(index);
                totalCost(index) = Inf;

                %find subsequent pixels, if appropriate
                for(iPixel = 2:intelligentAcquisition.numPixels)                    
                    [reward pixelProperties] = FindPixelReward(pixel{iFramePeek}, numDimensions, particle{1}, estUncertainty(iFrame-1),chosenPixels,pixelProperties,numDimensions+1,3*numDimensions,ones(2*numDimensions,1),-1);
                    rewardcostratio = reward./totalCost;
                    [tempratio index] = max(rewardcostratio);
                    if(tempratio > maxratio)
                        chosenPixels = [chosenPixels index];
                        totalCost = totalCost + cost(index);
                        totalCost(index) = Inf;
                    else
                        break
                    end
                end  
                
            %Choose pixels least likely to contain object
            case 7
            benefit = FindPixelProbability(pixel{iFramePeek}, numDimensions);
            benefit(isnan(benefit)) = 0;
            [temp order] = sort(benefit);
            chosenPixels = order(1:intelligentAcquisition.numPixels);

            %Choose all pixels
            otherwise
            chosenPixels = 1:(length(pixelBoundaries{iFrame})-1)^(numDimensions);
        end  
        
       %Choose the pixels to acquire for this frame, based on above method
        if(intelligentAcquisition.whereToAcquire ~= -1)
            pixel{iFramePeek} = ChoosePixel(pixel{iFramePeek}, chosenPixels,numDimensions); 
        end

       %Calculate the expected future uncertainty
       if(include(4))
           output.stoppingThreshold(iFramePeek,iFrame) = 0;
           for(iObject = 1:numObjects)
            indices = find(particle{iFramePeek}(:,particleMetaStart+1)==iObject);  
            [futureUncertainty(iFramePeek,iFrame,iObject)] = PredictUncertainty(particle{iFramePeek}(indices,:), numDimensions,3*numDimensions,numDimensions, pixel{iFramePeek}(indices,:));
            output.stoppingThreshold(iFramePeek,iFrame) = output.stoppingThreshold(iFramePeek,iFrame) + (estUncertainty(iFrame-1,iObject) - futureUncertainty(iFramePeek,iFrame,iObject));
           end
       end
   end
    

    
    %%%%%%%%WHEN TO STOP ACQUIRING%%%%%%%%
    
    %Initially assume we will not stop acquiring
    stopAcquiring = 0;
    switch intelligentAcquisition.stoppingCriteria
        %Stop when the estimated uncertainty drops below a threshold
        case 2
            if(estUncertainty(iFrame-1) < intelligentAcquisition.stoppingThreshold)
                stopAcquiring = 1;
            end
            
        %Stop when the expected reward for any frame over the next few frames is always
        %below a threshold
        case 4
            if(max(output.stoppingThreshold(:,iFrame)) < intelligentAcquisition.stoppingThreshold)
                stopAcquiring = 1;
            end
        
        %If none of these specified, just acquire until we run out of
        %frames
        otherwise
    end
    if(stopAcquiring == 1)
        break;
    end
    
    %Add the time cost of continued acquisition
    output.timeCost = output.timeCost + intelligentAcquisition.timeCost;
        
    %%%%%%%%WHEN TO ACQUIRE%%%%%%%%
    
    %Initially assume that we will acquire
    captureThis = 1;
    
    switch intelligentAcquisition.whenToAcquire
        %Acquire at a constant rate
        case 0
            if(acquireNow > 0)
                captureThis = 0;
            else
                acquireNow = acquireNow + 1;
            end
            acquireNow = acquireNow - 1/intelligentAcquisition.skipFrame;
        
        case 2
            %Acquire if benefit of acquiring frame A is higher than waiting
            %for B
            acquireNow = 0;
            acquireLater = 0;
            for(iObject = 1:numObjects)
                indices = find(particle{1}(:,particleMetaStart+1)==iObject);
                indices2 = find(particle{2}(:,particleMetaStart+1)==iObject);              
                acquireNow = acquireNow + estUncertainty(iFrame-1,iObject) - PredictUncertainty(particle{1}(indices,:), numDimensions,3*numDimensions,numDimensions, pixel{1}(indices,:));
                acquireLater = acquireLater + estUncertainty(iFrame-1,iObject) - PredictUncertainty(particle{2}(indices2,:), numDimensions,3*numDimensions,numDimensions, pixel{2}(indices2,:));
            end
            acquireNow = acquireNow / (1*intelligentAcquisition.photobleachingCost + (1+elapsedTime)*intelligentAcquisition.timeCost);
            acquireLater = acquireLater / (1*intelligentAcquisition.photobleachingCost + (2+elapsedTime)*intelligentAcquisition.timeCost);
            if(acquireLater > acquireNow)
                captureThis = 0;
                elapsedTime = elapsedTime + 1;
            else
                captureThis = 1;
                elapsedTime = 0;
            end
            
            
            %Acquire if the benefit of acquiring frames A & B is higher than
        %the benefit of just acquiring frame B
        case 3
            acquireBoth = 0;
            acquireLater = 0;
            for(iObject = 1:numObjects)
                indices = find(particle{1}(:,particleMetaStart+1)==iObject);
                indices2 = find(particle{2}(:,particleMetaStart+1)==iObject);
                numPixels = max(max(pixel{1}(indices,end)),max(pixel{2}(indices2,end)));
                pixelTrajectory = CreatePixelTrajectory(pixel{1}(indices,:),pixel{2}(indices2,:),numPixels);                            
                acquireBoth = acquireBoth + estUncertainty(iFrame-1,iObject) - PredictUncertainty(particle{1}(indices,:), numDimensions,3*numDimensions,numDimensions, pixelTrajectory);
                acquireLater = acquireLater + estUncertainty(iFrame-1,iObject) - PredictUncertainty(particle{2}(indices2,:), numDimensions,3*numDimensions,numDimensions, pixel{2}(indices2,:));
            end
            acquireBoth = acquireBoth / (2*intelligentAcquisition.photobleachingCost + (2+elapsedTime)*intelligentAcquisition.timeCost);
            acquireLater = acquireLater / (1*intelligentAcquisition.photobleachingCost + (2+elapsedTime)*intelligentAcquisition.timeCost);
            if(acquireLater > acquireBoth)
                captureThis = 0;
                elapsedTime = elapsedTime + 1;
            else
                captureThis = 1;
                elapsedTime = 0;
            end
            
        %Acquire randomly with the given inverse probability
        case 4
            if(rand(1)<(1/intelligentAcquisition.skipFrame))
                captureThis = 1;
            else
                captureThis = 0;
            end
                
    end
    
    
    %%%%%%%%UPDATING THE MODEL%%%%%%%%
    
    %Only update the model if we actually choose to acquire
    if(captureThis)  
        
        %Determine the pixel that each object lies in, assigning 0 if this
        %pixel was unacquired
        for(iObject = 1:numObjects)
            truePixel{iObject} = FindPixels(pixelBoundaries{iFrame}, position{iObject}(iFrame,:), numDimensions);
            if(intelligentAcquisition.whereToAcquire ~= -1)
                truePixel{iObject} = ChoosePixel(truePixel{iObject}, chosenPixels,numDimensions);
            end
        end
        
        %Overdetection - treat as poisson
        lastNumDetectedObjects = numDetectedObjects;
        numDetectedObjects = numObjects;
        
        %Mark the object as captured if it is not in pixel 0 (for single
        %object only)
        for iObject = 1:numObjects
            captured(iFrame,iObject) = truePixel{iObject}(1,numDimensions+1)>0;
        end
        
        %Increase the photobleaching cost if the object is captured (2D
        %imaging only - use LearnModel3D for 3D)
        if(captured(iFrame,1)==1)
            output.photobleachingCost = output.photobleachingCost + intelligentAcquisition.photobleachingCost;
        end

        %If detection is not perfect, assign objects to pixel 0 with a
        %given probability
        for(iObject = 1:numObjects)
            if(rand(1)<intelligentAcquisition.trueMisdetection)
                truePixel{iObject}(1,numDimensions+1) = 0;
            end
        end
        
        allIndices = [];

        %reset particle weights to 0 (assumes they were equal to start
        %with)
                   
        
        %Assign object IDs and weights to surviving particles
        for(iObject = 1:numObjects)
            %Find indices of particles in the same pixel as this object
            %Should we be giving less weight if multiple objects are in
            %this pixel? Seems like we should...but doesn't really make
            %a difference anyway since everything is halved equally
            indices2 = find(pixel{1}(:,numDimensions+1)==truePixel{iObject}(1,numDimensions+1));

            %Assign weights
            particle{1}(:,particleMetaStart+3) = 0;
            particle{1}(indices2,particleMetaStart+3) = 1 - intelligentAcquisition.perfectTracking;

            %Find correctly tracked particles and assign weights
            indices = find(particle{1}(indices2,particleMetaStart+1)==iObject);
            indices2 = indices2(indices);            
            particle{1}(indices2,particleMetaStart+3) = 1;

            %Sample for the object
            if(sum(particle{1}(:,particleMetaStart+3))==0)
                objectIndices{iObject} = (find(particle{1}(:,particleMetaStart+1)==iObject))';
            else
                objectIndices{iObject} = randsample(1:size(particle{1},1),numParticles,true,particle{1}(:,particleMetaStart+3));
            end
        end
               
        %Combine object indices
        allIndices = [];
        objectAssignments = [];
        for(iObject = 1:numDetectedObjects)
            allIndices = [allIndices; objectIndices{iObject}];
            objectAssignments = [objectAssignments; ones(length(objectIndices{iObject}),1)*iObject];               
        end
        particle{1} = particle{1}(allIndices,:);
        particle{1}(:,particleMetaStart+3) = 1/numParticles;

        %Reassign object IDs
        particle{1}(:,particleMetaStart+1) = objectAssignments;
        %for(iObject = 1:numDetectedObjects)
        %    particle{1}(((iObject-1)*numParticles+1):(iObject*numParticles),particleMetaStart+1)=iObject; 
        %end

        %Do a model jumble if desired
        if(intelligentAcquisition.modelBuilding == 1 || intelligentAcquisition.modelBuilding == 3 || intelligentAcquisition.modelBuilding == 5)
            newIndices = randperm(size(particle{1},1));
            particle{1}(:,modelIndices) = particle{1}(newIndices,modelIndices);
        end
    end
 

        %Roughening to protect against static parameters converging to the wrong value
        if(intelligentAcquisition.noiseMethod == 1)
            noiseMultiple = std(particle{1}(:,numDimensions+1:3*numDimensions));
        elseif(intelligentAcquisition.noiseMethod == 2)
            noiseMultiple = range(particle{1}(:,numDimensions+1:3*numDimensions));
        else
            noiseMultiple = ones(1,2*numDimensions);
        end
        particle{1}(:,numDimensions+1:3*numDimensions) = particle{1}(:,numDimensions+1:3*numDimensions) + randn(size(particle{1}(:,numDimensions+1:3*numDimensions)))*intelligentAcquisition.noise.*repmat(noiseMultiple,[size(particle{1},1) 1]);
        particle{1}(:,numDimensions+1:3*numDimensions) = max(particle{1}(:,numDimensions+1:3*numDimensions),0);
        particle{1}(:,numDimensions+1:3*numDimensions) = min(particle{1}(:,numDimensions+1:3*numDimensions),cellBoundary/2);
    
    
    %For each object, calculate its motion model prediction, its true
    %uncertainty, and its estimated uncertainty in this frame
    for(iObject = 1:numObjects)
        indices = find(particle{1}(:,particleMetaStart+1)==iObject);
        if(include(1))
            prediction(iFrame,iObject,1:2*numDimensions) = sum(particle{1}(indices,numDimensions+1:3*numDimensions).*repmat(particle{1}(indices,particleMetaStart+3),1,2*numDimensions),1)/sum(particle{1}(indices,particleMetaStart+3));
        end

        if(include(2))
            uncertainty(iFrame,iObject) = sum(sum(repmat(particle{1}(indices,particleMetaStart+3),1,2*numDimensions).*(particle{1}(indices,numDimensions+1:3*numDimensions)-repmat(truth{iObject}, [length(indices) 1])).^2))/sum(particle{1}(indices,particleMetaStart+3));
        end

        if(include(3))
            estUncertainty(iFrame,iObject) = 2*sum(var(particle{1}(indices,numDimensions+1:3*numDimensions)));
        end
    end %end object loop
end %end frame loop

output.priorParticles = priorParticles;